unit FileReaderAndWriterThreads;

interface

uses
  Classes,Semaphore;

const
  NUMBUFFERS = 2;
  BUFSIZE = 1024;

type
  TRWBuffer = record
    data : array[0..BUFSIZE-1] of char;
    size : integer;
  end;

  TReaderThread = class(TThread)
  private
    FFname    : string;
    FWriteSem : TPosixSemaphore;
    FReadSem  : TPosixSemaphore;
  protected
    procedure Execute; override;
    procedure ValidateProperties;
  public
    property Filename       : string          read FFname    write FFname;
    property ReadSemaphore  : TPosixSemaphore read FReadSem  write FReadSem;
    property WriteSemaphore : TPosixSemaphore read FWriteSem write FWriteSem;
  end;

  TWriterThread = class(TThread)
  private
    FFname    : string;
    FWriteSem : TPosixSemaphore;
    FReadSem  : TPosixSemaphore;
  protected
    procedure Execute; override;
    procedure ValidateProperties;
  public
    property Filename       : string          read FFname    write FFname;
    property ReadSemaphore  : TPosixSemaphore read FReadSem  write FReadSem;
    property WriteSemaphore : TPosixSemaphore read FWriteSem write FWriteSem;
  end;


implementation

uses
  SysUtils;

var
  theBuffer : array[0..NUMBUFFERS-1] of TRWBuffer;

{ TReaderThread }

procedure TReaderThread.Execute;
var
  i   : integer;
  InF : File;

begin
  try
    ValidateProperties;
    AssignFile(InF,FFname);
    Reset(InF,1);
    try
      i := 0;
      while true do
      begin
        // Wait for a read buffer to be available
        FReadSem.Wait;

        // read the next block of data
        BlockRead(InF,theBuffer[i].data,sizeof(theBuffer[i].data),theBuffer[i].size);

        if theBuffer[i].size = 0 then
        begin
          // we hit the end of the file
          FWriteSem.Post;
          break;
        end;

        inc(i);
        // treat this buffer like a circular one..
        if (i >= NUMBUFFERS) then
          i := 0;

        FWriteSem.Post;
      end;
    finally
      CloseFile(InF);
    end;
  except
    on e : exception do
    begin
      writeln('ReaderThread: Exception fired! ',e.message);
      Exit;
    end;
  end;
end;

procedure TReaderThread.ValidateProperties;
begin
  if FFname = '' then
    raise Exception.Create('the FileName property must be specified');

  if not FileExists(FFname) then
    raise Exception.CreateFmt('File %s does not exist!',[FFname]);

  if not Assigned(FReadSem) then
    raise Exception.Create('The ReadSemaphore property must be specified!');

  if not Assigned(FWriteSem) then
    raise Exception.Create('The WriteSemaphore property must be specified');
end;

{ TWriterThread }

procedure TWriterThread.Execute;
var
  i,recs : integer;
  OutF : File;

begin
  try
    ValidateProperties;
    AssignFile(OutF, FFname);
    Rewrite(OutF,1);
    try
      i := 0;
      while true do
      begin
        // wait for a buffer to write
        FWriteSem.Wait;

        // see if we hit the end of the file
        if theBuffer[i].size = 0 then
          break;

        BlockWrite(OutF, theBuffer[i].data, theBuffer[i].size, recs);

        inc(i);
        if i >= NUMBUFFERS then
          i := 0;

        // notify the readers there is slot available now
        FReadSem.Post;
      end;
    finally
      CloseFile(OutF);
    end;
  except
    on e : exception do
    begin
      writeln('WriterThread: Exception fired! ',e.message);
      Exit;
    end;
  end;
end;

procedure TWriterThread.ValidateProperties;
begin
  if FFname = '' then
    raise Exception.Create('the FileName property must be specified');

  if FileExists(FFname) then
    raise Exception.CreateFmt('File %s exists!',[FFname]);

  if not Assigned(FReadSem) then
    raise Exception.Create('The ReadSemaphore property must be specified!');

  if not Assigned(FWriteSem) then
    raise Exception.Create('The WriteSemaphore property must be specified');
end;

end.
