unit Queues;

interface

uses
  Classes, IPCBase, Mutex;

type
  TQueueType = (qtLifo, qtFifo);

  TThreadSafeQueue = class(TIPCBase)
  protected
    FCondFull  : TConditionVar;
    FCondEmpty : TConditionVar;
    FList      : TStrings;
    FQueueType : TQueueType;
    FMaxSize   : integer;
    FDone      : boolean;
  public
    constructor Create(QueueType : TQueueType; MaxSize : integer);
    destructor  Destroy; override;
    procedure   AddObject(obj : TObject);
    function    GetObject : TObject;
  end;

implementation

uses
  SysUtils;

{ TThreadSafeQueue }

procedure TThreadSafeQueue.AddObject(obj: TObject);
var
  bEmpty : boolean;

begin
  FCondFull.Mutex.Lock;
  try
    // If we're full, we need to wait
    while (FList.Count = FMaxsize) and (not FDone) do
      FCondFull.Wait;

    // The FDone variable indicates that Free has been
    // called. we don't want to do anything at this point
    // other than exit this method.
    if not FDone then
    begin
      // see if we will need to signal the Empty Condtion later
      bEmpty := FList.Count = 0;

      // add the object to the list
      case FQueueType of
        qtFifo : Flist.InsertObject(0,'',obj);
        qtLifo : FList.AddObject('',obj);
      else
        raise Exception.Create('AddObject: Unknown QueueType');
      end;

      // now signal the Empty condition, if necessary
      if bEmpty then
        FCondEmpty.Signal;
    end;

  finally
    FCondFull.Mutex.Unlock;
  end;
end;

constructor TThreadSafeQueue.Create(QueueType: TQueueType;
  MaxSize: integer);
begin
  inherited Create;
  FDone          := false;
  FQueueType     := QueueType;
  FMaxSize       := MaxSize;

  FCondFull      := TConditionVar.Create(nil);
  FCondEmpty     := TConditionVar.Create(FCondFull.Mutex);

  FList          := TStringList.Create;

  Flist.Capacity := FMaxSize;
end;

destructor TThreadSafeQueue.Destroy;
begin
  FDone := true;
  // now notify the condition variables
  DebugMessage('Before signaling other threads');
  FCondFull.Signal;
  FCondEmpty.Signal;
  DebugMessage('In TSQueue.Destroy');
  FCondFull.Free;
  DebugMessage('Before call to CondEmpty.Free');
  FCondEmpty.Free;
  DebugMessage('Before call to FList.Free');
  FList.Free;
  inherited;
end;

function TThreadSafeQueue.GetObject: TObject;
var
  bFull : boolean;
  slot  : integer;

begin
  Result := nil;
  FCondFull.Mutex.Lock;
  try
    // see if we have an empty queue
    while (FList.Count = 0) and (not FDone) do
      FCondEmpty.Wait;

    // The FDone variable indicates that Free has been
    // called. we don't want to do anything at this point
    // other than exit this method.
    if not FDone then
    begin
      // see if we need to notify the full later
      bFull := FList.Count = FMaxSize;

      case FQueueType of
        qtFifo : slot := 0;
        qtLifo : slot := FList.Count-1;
      else
        raise Exception.Create('GetObject: Unknown QueueType');
      end;

      // get the object and remove it from the list
      Result := FList.Objects[slot];
      FList.Delete(slot);

      // now, if the queue was previously full, we need to
      // notify the condition
      if bFull then
        FCondFull.Signal;
    end;

  finally
    FCondFull.Mutex.UnLock;
  end;
end;

end.
