unit SysVMsgQueue;

// Kylix object wrapper for System V Message Queues

interface

uses
   Libc, IPCBase;


type

   //
   // The following type is an example of what a message might look
   // like. The only requirement is the first field be an integer
   // that indicates the message type. It must be greater than zero.
   // Anything after the first field can be whatever is needed and
   // is not limited to character data types.
   //
   TMQExampleMsg = packed record
      MessageType : integer;                // required, > 0
      MessageText : array[0..1] of char;    // whatever you require
   end;

   TMQFilePermission = (fpOwnerRead, fpOwnerWrite,
                        fpGroupRead, fpGroupWrite,
                        fpOtherRead, fpOtherWrite);

   TMQFilePermissions = set of TMQFilePermission;

   TSystemVMQ = class(TIPCBase)
     private
       FMsgKey      : key_t;    // key needed for Message Queue
       FMsgID       : integer;  // refers to the queue created by msgget
       FOpened      : boolean;
       FPrivateKey  : boolean;
       FPathName    : string;
       FIPCKey      : integer;
       FPermissions : TMQFilePermissions;
       FOpenExist   : boolean;
       FBlockRead   : boolean;
       FBlockWrite  : boolean;

       procedure SetIPCKey(const value : integer);
       function  FilePermissionsToBits : integer;
     protected
       procedure ValidateProperties;

     public
       constructor Create;
       destructor  Destroy; override;

       procedure   Open;
       function    Write(buffer : Pointer; MessageLength : integer) : boolean;
       function    Read(buffer : Pointer;
                        buflen : integer;
                        MessageType : integer;
                        var NumRead : integer) : boolean;
       procedure   Close;

       property    PrivateKey   : boolean            read FPrivateKey  write FPrivateKey;
       property    PathName     : string             read FPathName    write FPathName;
       property    IPCKey       : integer            read FIPCKey      write SetIPCKey;
       property    Permissions  : TMQFilePermissions read FPermissions write FPermissions;
       property    OpenExisting : boolean            read FOpenExist   write FOpenExist    default false;
       property    BlockOnRead  : boolean            read FBlockRead   write FBlockRead    default true;
       property    BlockOnWrite : boolean            read FBlockWrite  write FBlockWrite   default true;
   end;

implementation

uses
   SysUtils, SysVMsgQueueConstants;

constructor TSystemVMQ.Create;
begin
  inherited;
  FPrivateKey  := false;
  FOpened      := false;
  FPathName    := '';
  FIPCKey      := 0;
  FPermissions := [];
  FOpenExist   := false;   // default it to create mq if it doesnt exist.
  FBlockRead   := true;    // default is to block
  FBlockWrite  := true;    // default is to block
  FMsgID       := 0;
end;

destructor TSystemVMQ.Destroy;
begin
  // clean up
  inherited;
end;

procedure TSystemVMQ.Open;
var
  ret,oflag : integer;
  tmpFile : file;

begin
  DebugMessage('Opening..');

  ValidateProperties;

  if FPrivateKey then
     FMsgKey := IPC_PRIVATE
  else
  begin
     // make sure the file exists..
     if not FileExists(FPathName) then
     begin
        // create a file if we can..
        Assign(tmpFile,FPathName);
        Rewrite(tmpFile);
        CloseFile(tmpFile);
     end;
     DebugMessage('Before call to ftok');
     FMsgKey := ftok(PChar(FPathName),FIPCKey);
     if FMsgKey = LIBC_FAILURE then
     begin
        DebugMessage('ftok failed error is ' + IntToStr(GetLastError));
        ErrorMessage('Unable to ftok message queue');
     end;
  end;

  oflag := IPC_CREAT;
  // if we want to only open it, then mask in the IPC_EXCL bit(s)
  if FOpenExist then
     oflag := oflag or IPC_EXCL;

  // now get the permission bits and or them to oflag
  oflag := oflag or FilePermissionsToBits;

  DebugMessage('Before call to msgget');

  // do the open
  ret := msgget(FMsgKey, oflag);
  if (ret = LIBC_FAILURE) then
     ErrorMessage('msgget failed');

  FOpened := true;
end;

//
// Buffer points to a message structure similar to TMQExampleMsg -
// that is the first field is an integer that must be greater than
// zero.
//
// MessageLength is the size of the MESSSAGE not the entire record.
// The easiest way to calculate this is to use the following:
//
// MessageLength := sizeof(TMQExampleMsg) - sizeof(integer);
//

function TSystemVMQ.Write(buffer : Pointer;
                          MessageLength : integer) : boolean;
var
  ret,flag : integer;

begin
  if not FOpened then
  begin
     DebugMessage('Message Queue is not opened! (write)');
     Result := false;
  end
  else
  begin
     flag := 0;
     if not FBlockWrite then
        flag := IPC_NOWAIT;

     Result := true;
     ret := msgsnd(FMsgID, buffer, MessageLength, flag);
     if (ret = LIBC_FAILURE) then
     begin
        DebugMessage('msgsnd failed ' + IntToStr(GetLastError));
        Result := false;
     end;
  end;
end;

//
// Buffer - is where the message will be stored. It also points to where
//          the MessageType field.
//
// buflen - specifies the length of the buffer
//
// MessageType - indicates what message is needed:
//   MessageType = 0 - first message in the queue is returned (the oldest)
//
//   MessageType < 0 - the first message with the lowest type that is less
//                     than or equal to the absolute value of MessageType
//
//   MessageType > 0 - the first message that matches the type specified
//
// NumRead - returns the number of bytes of data read excluding the
//           initial field of MessageType.

function TSystemVMQ.Read(buffer : Pointer;
                         buflen : integer;
                         MessageType : integer;
                         var NumRead : integer) : boolean;
var
   ret,flag : integer;

begin
  if not FOpened then
  begin
     DebugMessage('Message Queue is not opened! (read)');
     Result := false;
  end
  else
  begin
     flag := 0;
     if not FBlockRead then
        flag := IPC_NOWAIT;

     DebugMessage('Before call to msgrcv..');

     Result := true;
     ret := msgrcv(FMsgID, buffer, buflen, MessageType, flag);
     if ret = LIBC_FAILURE then
     begin
        DebugMessage('msgrcv failed error ' + IntToStr(GetLastError));
        Result := false;
     end;
  end;
end;

//
// Note The msgctl api does more than just remove the message queue.
//

procedure TSystemVMQ.Close;
var
  ret : integer;

begin
  if FOpened then
  begin
    ret := msgctl(FMsgID, IPC_RMID, nil);
    if ret = LIBC_FAILURE then
       ErrorMessage('Error closing msq queue (msgctl)');
    FOpened := false;
    // clean up the file if we can..
    DeleteFile(FPathName);
  end;
end;


procedure TSystemVMQ.ValidateProperties;
begin
  // either FPrivateKey needs to be set
  // or both FPathName and FIPCKey needs to be set
  if (not FPrivateKey) and (FPathName <> '') and (FIPCKey > 0) then
     ErrorMessage('Either the PrivateKey property needs' +
                  ' to be set or both PathName and IPCKey ');
end;

procedure TSystemVMQ.SetIPCKey(const value : integer);
begin
  if value > 0 then
     FIPCKey := value
  else
     ErrorMessage('IPC Key must be a positive integer');
end;

function TSystemVMQ.FilePermissionsToBits : integer;
begin
   Result := 0;

   if fpOwnerRead  in FPermissions then
      Result := Result or USR_MSG_R;

   if fpOwnerWrite in FPermissions then
      Result := Result or USR_MSG_W;

   if fpGroupRead  in FPermissions then
      Result := Result or GRP_MSG_R;

   if fpGroupWrite in FPermissions then
      Result := Result or GRP_MSG_W;

   if fpOtherRead  in FPermissions then
      Result := Result or WRLD_MSG_R;

   if fpOtherWrite in FPermissions then
      Result := Result or WRLD_MSG_W;
end;

end.
